"""
Kwenta 🧮

"""

from kwenta import Kwenta
from loguru import logger

from .client import DexClient


class KwentaHandler(DexClient):
    """
    A DexClient using kwenta-python library

    More info
    https://github.com/Kwenta/kwenta-python-sdk

    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initialize the client

        """
        super().__init__(**kwargs)
        if self.rpc is None or self.w3 is None:
            return
        self.client = Kwenta(
            network_id=int(self.w3.net.version),
            provider_rpc=self.rpc,
            wallet_address=self.wallet_address,
            private_key=self.private_key,
        )

    async def get_quote(
        self,
        buy_address=None,
        buy_symbol=None,
        sell_address=None,
        sell_symbol=None,
        amount=1,
    ):
        """
        Retrieves a quote for a given symbol.

        Args:
            buy_address (str, optional): The buy address. Defaults to None.
            symbol (str, optional): The symbol to retrieve the quote for.
            Defaults to None.
            amount (int, optional): The amount of the asset to retrieve the quote for.
            Defaults to 1.

        Returns:
            float: The quote for the specified symbol.

        Raises:
            Exception: If an error occurs during the retrieval process.
        """
        try:
            buy_token = await self.resolve_token(
                address_or_symbol=buy_address
                or buy_symbol
                or self.trading_asset_address
            )
            sell_token = await self.resolve_token(
                address_or_symbol=sell_address or sell_symbol
            )

            if not buy_token or not sell_token:
                return "⚠️ Buy or sell token not found"

            market = self.client.markets[f"{sell_token.symbol}"]
            logger.info("market: {}", market)
            quote = self.client.get_current_asset_price(sell_token.symbol)
            logger.info("quote: {}", quote)
            return quote or "⚠️ Quote failed"
        except Exception as error:
            logger.error("Quote failed {}", error)
            return f"⚠️ {error}"

    async def make_swap(self, sell_address, buy_address, amount):
        pass

        # symbol = self.contract_utils.get_token_symbol(buy_address)
        # logger.debug(f"Symbol: {symbol}\n")
        # # get the market info for the asset
        # market = self.client.markets[symbol]
        # logger.debug(f"Market: {market}\n")

        # # check margin balance
        # margin_balance_before = self.client.get_accessible_margin(symbol)
        # logger.debug(f"Starting margin balance: {margin_balance_before}\n")

        # # transfer margin to the market
        # transfer_margin = self.client.transfer_margin(symbol, 100, execute_now=True)
        # logger.debug(f"Transfer tx: {transfer_margin}\n")

        # # wait for the transfer to be mined
        # self.client.web3.eth.wait_for_transaction_receipt(transfer_margin)
        # logger.debug("Transfer transaction confirmed\n")
        # await asyncio.sleep(10)

        # # check margin balance again
        # margin_balance_after = self.client.get_accessible_margin(symbol)
        # logger.debug(f"Ending margin balance: {margin_balance_after}\n")

        # # submit an order
        # open_position = self.client.modify_position(
        # symbol,
        # size_delta=0.5,
        # execute_now=True)
        # logger.debug(f"Open position tx: {open_position}\n")

        # # wait for the transfer to be mined
        # self.client.web3.eth.wait_for_transaction_receipt(open_position)
        # logger.debug("Order transaction confirmed\n")
        # await asyncio.sleep(2)

        # # check the order status
        # order_before = self.client.check_delayed_orders(symbol)
        # logger.debug(f"Order before: {order_before}\n")
        # await asyncio.sleep(20)

        # # check the open position
        # position = self.client.get_current_position(symbol)
        # logger.debug(f"Position: {position}\n")

        # # check the order status again
        # order_after = self.client.check_delayed_orders(symbol)
        # logger.debug(f"Order after: {order_after}\n")



"""
 DEX SWAP
🛠️ W3 UTILS
"""

import aiohttp
from loguru import logger


async def fetch_url(url, params=None, headers=None):
    """
    Asynchronously gets a url payload
    and returns the response.

    Args:
        url (str): The url to get.
        params (dict, optional): The params to send. Defaults to None.
        headers (dict, optional): The headers to send. Defaults to None.

    Returns:
        dict or None: The response or None if an
        error occurs or the response is too large.

    """
    max_response_size = 10 * 1024 * 1024  # 10 MB
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, params=params, headers=headers, timeout=20
            ) as response:
                if response.status == 200:
                    if (
                        response.content_length
                        and response.content_length > max_response_size
                    ):
                        logger.warning("Response content is too large to process.")
                        return None
                    return await response.json(content_type=None)
                logger.warning(f"Received non-200 status code: {response.status}")
    except Exception as error:
        logger.error(f"Unexpected error occurred: {error}")
    return None



"""
 DEX SWAP
🪙 TOKEN
"""

from loguru import logger

from dxsp.utils.utils import fetch_url


class Token:
    """

    Class Token to interact with web3 token contract

    Args:
        w3: An instance of the web3 library.
        address: The address of the token contract.
        block_explorer_url: The URL of the block explorer for the token.
        block_explorer_api: The API endpoint of the block explorer for the token.
        symbol: The symbol of the token.

    Returns:
        Token

    Methods:
        __init__: Initializes an instance of the class.
        fetch_data: Retrieves data for the token.
        get_token_abi: Retrieves the token abi.
        get_token_contract: Retrieves the token contract.
        get_contract_function: Retrieves the contract functions by name.
        get_token_balance: Retrieves the token balance.
        get_token_symbol: Retrieves the token symbol.
        get_token_name: Retrieves the token name.
        get_token_decimals: Retrieves the token decimals.

    """

    def __init__(self, **kwargs):
        """
        Initializes an instance of the class.

        :param w3: An instance of the web3 library.
        :type w3: object

        :param address: The address of the token contract.
        :type address: str

        :param block_explorer_url: The URL of the block explorer for the token.
        :type block_explorer_url: str

        :param block_explorer_api: The API endpoint of the block explorer for the token.
        :type block_explorer_api: str

        :param symbol: The symbol of the token.
        :type symbol: str
        """
        try:
            get = kwargs.get
            logger.debug("initializing token")
            self.w3 = get("w3", None)
            self.address = self.w3.to_checksum_address(get("address", None))
            self.symbol = get("symbol", None)
            self.headers = get("headers", None)
            self.abi_url = get("abi_url", None)
            self.block_explorer_url = get("block_explorer_url", None)
            self.block_explorer_api = get("block_explorer_api", None)
            self.decimals = None
            self.name = None
            logger.debug(
                "token initialized symbol {} address {}", self.symbol, self.address
            )
        except Exception as error:
            logger.error("token error {}", error)

    async def fetch_data(self) -> None:
        """
        Retrieves data for the token.
        """
        logger.debug("fetch token data")
        self.contract = await self.get_token_contract() or None
        self.decimals = await self.get_token_decimals() or 18
        self.symbol = await self.get_token_symbol() or None
        self.name = await self.get_token_name() or None
        logger.debug("{} - token data {}", self.symbol, self.address)

    async def get_token_abi(self, address=None):
        """
        Retrieves the ABI (Application Binary Interface)
        of a token contract at the given address.
        First, it tries to fetch the ABI from the block explorer
        if the block explorer url and api key are provided.
        If not, it tries to fetch the ABI from the dex ERC20 ABI URL

        Args:
            address (str, optional): The address of the token contract.
            If not provided, the address associated with the
             contract instance is used. Defaults to None.

        Returns:
            str: The ABI of the token contract, if successful.
            None if the request fails or the contract does not have an ABI.
        """
        if self.block_explorer_url and self.block_explorer_api:
            if address is None:
                address = self.address
            params = {
                "module": "contract",
                "action": "getabi",
                "address": address,
                "apikey": self.block_explorer_api,
            }
            resp = await fetch_url(
                url=self.block_explorer_url, headers=self.headers, params=params
            )
            if resp:
                return resp["result"] if resp["status"] == "1" else None

        logger.warning("No block explorer url or api key provided")
        return await fetch_url(url=self.abi_url)

    async def get_token_contract(self):
        """
        Retrieves the token contract.

        :return: The token contract.
        """
        self.abi = await self.get_token_abi()
        logger.debug("token abi {}", self.abi)
        if self.abi is None:
            return None
        contract = self.w3.eth.contract(address=self.address, abi=self.abi)
        return contract

    # def get_contract_function(self, contract, func_name: str):
    #     """
    #     Get the contract function by name.

    #     Args:
    #         contract: The contract object.
    #         func_name (str): The name of the function.

    #     Returns:
    #         bool: True if the function exists
    #         in the contract, False otherwise.
    #     """
    #     return func_name in dir(contract.functions)

    async def get_token_balance(self, wallet_address):
        """
        Get the balance of a token for a given wallet address.

        Args:
            wallet_address (str): The wallet
            address to check the balance for.

        Returns:
            float: The balance of the token in ether.

        Raises:
            None
        """
        contract = await self.get_token_contract()
        if not contract or contract.functions is None:
            logger.warning("No Balance")
            return 0
        balance = contract.functions.balanceOf(wallet_address).call()
        return round(self.w3.from_wei(balance, "ether"), 5) or 0

    async def get_token_symbol(self):
        """
        Retrieves the symbol of the token.

        Returns:
            str: The symbol of the token.
        """
        contract = await self.get_token_contract()
        if contract:
            return contract.functions.symbol().call()

    async def get_token_name(self):
        """
        Get the name of the token.

        Args:
            self: The object itself.

        Returns:
            The name of the token as a string.
        """
        contract = await self.get_token_contract()
        if contract:
            return contract.functions.name().call()

    async def get_token_decimals(self):
        """
        Get the number of decimal places for the token.

        Returns:
            int: The number of decimal places for the token.
        """
        if self.decimals is not None:
            return self.decimals
        contract = await self.get_token_contract()
        return contract.functions.decimals().call() if contract else 18



"""
 DEX SWAP Main
"""

import importlib

from loguru import logger

from dxsp import __version__
from dxsp.config import settings


class DexSwap:
    """
    class to build a DexSwap Object
    use to interact with the dex protocol

     Returns:
         DexSwap

     Methods:
        _create_client()
        get_info
        get_balances
        get_positions
        get_pnl
        get_quotes()
        submit_order

    """

    def __init__(self):
        """
        Initializes the class instance by creating and appending clients
        based on the configuration in `settings.cex`.

        Checks if the module is enabled by looking at `settings.myllm_enabled`.
        If the module is disabled, no clients will be created.

        Creates a mapping of library names to client classes.
        This mapping is used to create new clients based on the configuration.

        If a client's configuration exists in `settings.cex_enabled` and is truthy,
        it will be created.
        Clients are not created if their name is "template" or empty string.

        If a client is successfully created, it is appended to the `clients` list.

        If a client fails to be created, a message is logged with the name of the
        client and the error that occurred.

        Parameters:
            None

        Returns:
            None
        """
        # Check if the module is enabled
        self.enabled = settings.dxsp_enabled

        # Create a mapping of library names to client classes
        self.client_classes = self.get_all_client_classes()
        # logger.debug("client_classes available {}", self.client_classes)

        if not self.enabled:
            logger.info("Module is disabled. No Client will be created.")
            return
        self.clients = []
        # Create a client for each client in settings.dex
        for name, client_config in settings.dex.items():
            if (
                # Skip empty client configs
                client_config is None
                # Skip non-dict client configs
                or not isinstance(client_config, dict)
                # Skip template and empty string client names
                or name in ["", "template"]
                # Skip disabled clients
                or not client_config.get("enabled")
            ):
                continue

            # Create the client
            logger.debug("Creating client {}", name)
            client = self._create_client(**client_config, name=name)
            # If the client has a valid client attribute, append it to the list
            if client and getattr(client, "client", None):
                self.clients.append(client)

        # Log the number of clients that were created
        logger.info(f"Loaded {len(self.clients)} clients")
        if not self.clients:
            logger.warning(
                "No Client were created. Check your settings or disable the module."
            )
            return None

    def _create_client(self, **kwargs):
        """
        Create a client based on the given protocol.

        This function takes in a dictionary of keyword arguments, `kwargs`,
        containing the necessary information to create a client. The required
        key in `kwargs` is "library", which specifies the protocol to use for
        communication with the Dex. The value of "library" must match one of the
        libraries supported by DXSP.

        This function retrieves the class used to create the client based on the
        value of "library" from the mapping of library names to client classes
        stored in `self.client_classes`. If the value of "library" does not
        match any of the libraries supported, the function returns None.

        If the class used to create the client is found, the function creates a
        new instance of the class using the keyword arguments in `kwargs` and
        returns it.

        The function returns a client object based on the specified protocol
        or None if the library is not supported.

        Parameters:
            **kwargs (dict): A dictionary of keyword arguments containing the
            necessary information for creating the client. The required key is
            "library".

        Returns:
            A client object based on the specified protocol or None if the
            library is not supported.

        """
        library = kwargs.get("library") or kwargs.get("protocol") or "uniswap"
        return self.client_classes.get(f"{library.capitalize()}Handler", None).__call__(
            **kwargs
        )

    def get_all_client_classes(self):
        """
        Retrieves all client classes from the `dxsp.handler` module.

        This function imports the `dxsp.handler` module and retrieves
        all the classes defined in it.

        The function returns a dictionary where the keys are the
        names of the classes and the values are the corresponding
        class objects.

        Returns:
            dict: A dictionary containing all the client classes
            from the `dxsp.handler` module.
        """
        provider_module = importlib.import_module("dxsp.handler")
        return {
            name: cls
            for name, cls in provider_module.__dict__.items()
            if isinstance(cls, type)
        }

    async def get_info(self):
        """
        Retrieves information about the exchange
        and the account.

        :return: A formatted string containing
        the exchange name and the account information.
        :rtype: str
        """
        version_info = f"ℹ️ {type(self).__name__} {__version__}\n"
        client_info = "".join(
            f"💱 {client.name} {client.account_number}\n" for client in self.clients
        )
        return version_info + client_info.strip()

    async def get_balances(self):
        """
        Retrieves the account balance.

        :return: The account balance.
        :rtype: float
        """
        _info = []
        for client in self.clients:
            _info.append(f"\n{await client.get_account_balance()}")
        return "\n".join(_info)

    async def get_positions(self):
        """
        Retrieves the account position.

        :return: The account position.
        :rtype: AccountPosition
        """
        _info = []
        for client in self.clients:
            _info.append(f"{client.name}:\n{await client.get_account_position()}")
        return "\n".join(_info)

    async def get_pnls(self, **kwargs):
        """
        Retrieves the account position.

        :return: The account position.
        :rtype: AccountPosition
        """
        _info = []
        for client in self.clients:
            client_name = f"{client.name}:\n"
            account_pnl = await client.get_account_pnl(
                period=kwargs.get("period", None)
            )
            client_info = f"{client_name}{account_pnl}"
            _info.append(client_info)
        return "\n".join(_info)

    async def get_quotes(self, symbol=None, address=None):
        """
        gets a quote for a token

        Args:
            sell_token (str): The sell token.

        Returns:
            str: The quote with the trading symbol

        """
        _info = []
        for client in self.clients:
            quote = await client.get_quote(sell_symbol=symbol, sell_address=address)
            client_info = f"{client.name}: {quote}"
            _info.append(client_info)
            logger.debug("Retrieved quote for {}: {}", client.name, quote)

        logger.debug("All quotes: {}", " | ".join(_info))
        return "\n".join(_info)

    async def submit_order(self, order_params):
        """
        Execute an order function.

        Args:
            order_params (dict): The order parameters.

        Returns:
            str: The trade confirmation

        """
        _order = []
        for client in self.clients:
            try:
                action = order_params.get("action")
                instrument = await client.replace_instrument(
                    order_params.get("instrument")
                )
                quantity = order_params.get("quantity", 1)
                sell_token, buy_token = (
                    (client.trading_asset_address, instrument)
                    if action == "BUY"
                    else (instrument, client.trading_asset_address)
                )
                order = await client.get_swap(sell_token, buy_token, quantity)
                if "⚠️" in order:
                    _order.append(f"{client.name}:\n{str(order)}\n")
                else:
                    order_info = (
                        f"{client.name}:\n⬇️ {instrument}"
                        if action == "SELL"
                        else f"{client.name}:\n⬆️ {instrument}\n"
                    )
                    order_info += order
                    _order.append(order_info)
            except Exception as error:
                logger.error("Error submitting order for {}: {}", client.name, error)
                _order.append(f"⚠️ {client.name} {error}")
        return "\n".join(_order)



from .kwenta import KwentaHandler

# from .oneinch import OneinchHandler
from .uniswap import UniswapHandler
from .zerox import ZeroxHandler

__all__ = ["UniswapHandler", "ZeroxHandler", "KwentaHandler"]



from dxsp.utils.account_utils import AccountUtils
from dxsp.utils.contract_utils import ContractUtils
from dxsp.utils.token_utils import Token

__all__ = [
    "AccountUtils",
    "ContractUtils",
    "Token",
]
